/* Class.java -- Representation of a Java class.
   Copyright (C) 1998, 1999, 2000, 2002, 2003, 2004

This file is part of GNU Classpath.

GNU Classpath is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Classpath is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Classpath; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version. */

package java.lang;

import java.io.InputStream;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.security.AllPermission;
import java.security.Permissions;
import java.security.ProtectionDomain;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import gnu.java.lang.ClassHelper;
import org.ovmj.java.Opaque;

/**
 * A Class represents a Java type.  There will never be multiple Class
 * objects with identical names and ClassLoaders. Primitive types, array
 * types, and void also have a Class object.
 *
 * <p>Arrays with identical type and number of dimensions share the same
 * class.  The name of an
 * array class is <code>[&lt;signature format&gt;;</code> ... for example,
 * String[]'s class is <code>[Ljava.lang.String;</code>. boolean, byte,
 * short, char, int, long, float and double have the "type name" of
 * Z,B,S,C,I,J,F,D for the purposes of array classes.  If it's a
 * multidimensioned array, the same principle applies:
 * <code>int[][][]</code> == <code>[[[I</code>.
 *
 * <p>There is no public constructor - Class objects are obtained only through
 * the virtual machine, as defined in ClassLoaders.
 *
 *
 * @serialData Class objects serialize specially:
 * <code>TC_CLASS ClassDescriptor</code>. For more serialization information,
 * see {@link ObjectStreamClass}.
 *
 * @author John Keiser
 * @author Eric Blake <ebb9@email.byu.edu>
 * @since 1.0
 * @see ClassLoader
 */
public final class Class implements Serializable {

    /** Opaque reference to the associated OVM Type object */
    /* package access*/final Opaque vmType; 


    /** The name of the Class represented by this. (lazily initialized) */
    private volatile String name;


    private static final int CLASS     = 0x00;
    private static final int INTERFACE = 0x01;
    private static final int ARRAY     = 0x02;
    private static final int PRIMITIVE = 0x03;

    /** What kind of Class are we */
    private final int kind;

    /** Modifiers for this class */
    private final int modifiers;

    /**
     * Compatible with JDK 1.0+.
     */
    private static final long serialVersionUID = 3206093459760846163L;

    /** The class signers. */
    private Object[] signers = null;

    /** The class protection domain. */
    private ProtectionDomain pd = null;

    /** The unknown protection domain. */
    private static volatile ProtectionDomain unknownProtectionDomain;

    static {
        Permissions permissions = new Permissions();
        permissions.add(new AllPermission());
        unknownProtectionDomain = new ProtectionDomain(null, permissions);
    }

    /* For caching of Method, Constructor, and Field objects, used for 
       reflection. Only heap and immortal allocated objects are cached.
       Scope usage of these methods will reallocate each time.
    */
    private volatile transient Method      publicMethods[];
    private volatile transient Method      declaredMethods[];
    private volatile transient Constructor publicConstructors[];
    private volatile transient Constructor declaredConstructors[];
    private volatile transient Field       publicFields[];
    private volatile transient Field       declaredFields[];

    /** stores the methods provided by this class
     * @see #initMethodHash */
    private HashMap methodHashMap;
    /** stores the constructorsprovided by this class
     * @see #initMethodHash */
    private HashMap constructorHashMap;
    
    /* just to avoid having many copies of the "<init>" string around */
    private static final String constructorName = "<init>";


    /**
     * Called by the OVM to instantiate a Class object for a given OVM Type
     * (where a Type is a type-name, type-context pair).
     * This constructor pre-initializes most of the static information about
     * a class.
     * @param vmType Opaque reference to our associated OVM Type object.
     * @param isInterface indicates if this Class object represents an
     * interface
     * @param isArray indicates if this Class object represents an array
     * @param isPrimitive indicates if this Class object represents a primitive
     * type
     * @param modifiers the modifiers for the class represented by this Class
     * object
     */
    private Class(Opaque vmType, boolean isInterface, boolean isArray,
                  boolean isPrimitive, int modifiers) {
	this.vmType = vmType;
        // the kinds of class are mutually exclusive
	kind = (isInterface ? INTERFACE :
		isArray     ? ARRAY :
		isPrimitive ? PRIMITIVE :
		CLASS);
	this.modifiers = modifiers;
    }


    // class objects are singletons so we use identity only
    public boolean equals(Class other) {
	return (this == other);
    }

    /**
     * Return the human-readable form of this Object.  For an object, this
     * is either "interface " or "class " followed by <code>getName()</code>,
     * for primitive types and void it is just <code>getName()</code>.
     *
     * @return the human-readable form of this Object
     */
    public String toString() {
        if (isPrimitive())
            return getName();
        return (isInterface() ? "interface " : "class ") + getName();
    }

    /**
     * Use the classloader of the current class to load, link, and initialize
     * a class. This is equivalent to your code calling
     * <code>Class.forName(name, true, getClass().getClassLoader())</code>.
     *
     * @param name the name of the class to find
     * @return the Class object representing the class
     * @throws ClassNotFoundException if the class was not found by the
     *         classloader
     * @throws LinkageError if linking the class fails
     * @throws ExceptionInInitializerError if the class loads, but an exception
     *         occurs during initialization
     */
    public static Class forName(String name)
        throws ClassNotFoundException {
        if (name == null) {
            throw new NullPointerException("no class name specified");
        }
        return forName(name, true,
                       VMSecurityManager.getClassContext()[1].getClassLoader());
      }
    
    /**
     * Use the specified classloader to load and link a class. If the loader
     * is null, this uses the bootstrap class loader (provide the security
     * check succeeds). Unfortunately, this method cannot be used to obtain
     * the Class objects for primitive types or for void, you have to use
     * the fields in the appropriate java.lang wrapper classes.
     *
     * <p>Calls <code>classloader.loadclass(name, initialize)</code>.
     *
     * @param name the name of the class to find
     * @param initialize whether or not to initialize the class at this time
     * @param classloader the classloader to use to find the class; null means
     *        to use the bootstrap class loader
     * @throws ClassNotFoundException if the class was not found by the
     *         classloader
     * @throws LinkageError if linking the class fails
     * @throws ExceptionInInitializerError if the class loads, but an exception
     *         occurs during initialization
     * @throws SecurityException if the <code>classloader</code> argument
     *         is <code>null</code> and the caller does not have the
     *         <code>RuntimePermission("getClassLoader")</code> permission
     * @see ClassLoader
     * @since 1.2
     */
    public static Class forName(String name, boolean initialize,
                                ClassLoader classloader)
        throws ClassNotFoundException
    {
	if (classloader == null) {
	    // Check if we may access the bootstrap classloader
	    SecurityManager sm = System.getSecurityManager();
	    if (sm != null) {
		// Get the calling class and classloader
		Class c = VMSecurityManager.getClassContext()[1];
		ClassLoader cl = c.getClassLoader();
		if (cl != null)
		    sm.checkPermission(new RuntimePermission("getClassLoader"));
	    }
	    classloader = VMClassLoader.bootClassLoader;
	}
	try {
	    Class c = LibraryImports.initiateLoading(classloader,
						     name.intern());
	    if (initialize)
		LibraryImports.initializeClass(c.vmType);
	    return c;
	} catch (NoClassDefFoundError e) {
	    throw new ClassNotFoundException(name, e);
	}
    }

    /**
     * Get a new instance of this class by calling the no-argument constructor.
     * The class is initialized if it has not been already. A security check
     * may be performed, with <code>checkMemberAccess(this, Member.PUBLIC)</code>
     * as well as <code>checkPackageAccess</code> both having to succeed.
     *
     * <p><B>OVM NOTE:</b> Despite what the throws documentation says
     * regarding throwing <tt>InstantiationException</tt> if &quot;an
     * exception occurred during the constructor&quot;, in fact both the
     * JDK and Classpath now &quot;magically&quot; throw any exception
     * that was thrown by the constructor, including checked exceptions that
     * are not declared on this method. This is a stupid thing to do but the
     * JDK now seems to have entrenched this behaviour, despite it being
     * reported as a bug for years. The best answer is to consider this method
     * deprecated and always use the <tt>Constructor</tt> object directly.
     *
     * @return a new instance of this class
     * @throws InstantiationException if there is not a no-arg constructor
     *         for this class, including interfaces, abstract classes, arrays,
     *         primitive types, and void; or if an exception occurred during
     *         the constructor
     * @throws IllegalAccessException if you are not allowed to access the
     *         no-arg constructor because of scoping reasons
     * @throws SecurityException if the security check fails
     * @throws ExceptionInInitializerError if class initialization caused by
     *         this call fails with an exception
     */
    public Object newInstance()
        throws InstantiationException, IllegalAccessException {
        try {
            // The constructor need not be public but should be accessible.
            // FIXME: Currently we don't perform accessibility checks.
            return getDeclaredConstructor(null).newInstance(null);
        }
        catch (IllegalArgumentException e) {
            throw (Error) new InternalError("Should not happen").initCause(e);
        }
        catch (InvocationTargetException e) {
            throw LibraryImports.uncheckedThrow(e.getCause());
        }
        catch (NoSuchMethodException e) {
            throw (InstantiationException) 
                new InstantiationException(e.toString()).initCause(e);
        }
    }
    
    /**
     * Discover whether an Object is an instance of this Class.  Think of it
     * as almost like <code>o instanceof (this class)</code>.
     *
     * @param o the Object to check
     * @return whether o is an instance of this class
     * @since 1.1
     */
    public boolean isInstance(Object o){
        // not allowed to throw NPE
	return o == null ? false : isAssignableFrom(o.getClass());
    }

    /**
     * Discover whether an instance of the Class parameter would be an
     * instance of this Class as well.  Think of doing
     * <code>isInstance(c.newInstance())</code> or even
     * <code>c.newInstance() instanceof (this class)</code>. While this
     * checks widening conversions for objects, it must be exact for primitive
     * types.
     *
     * @param c the class to check
     * @return whether an instance of c would be an instance of this class
     *         as well
     * @throws NullPointerException if c is null
     * @since 1.1
     */
    public boolean isAssignableFrom(Class c) {
        // no need for null check as we dereference c
	return LibraryImports.isSubtypeOf(c.vmType, vmType);
    }
    
    /**
     * Check whether this class is an interface or not.  Array types are not
     * interfaces.
     *
     * @return whether this class is an interface or not
     */
    public boolean isInterface() {
	return kind == INTERFACE;
    }
    
    /**
     * Return whether this class is an array type.
     *
     * @return whether this class is an array type
     * @since 1.1
     */
    public boolean isArray() {
        return kind == ARRAY;
    }
    
    /**
     * Return whether this class is a primitive type.  A primitive type class
     * is a class representing a kind of "placeholder" for the various
     * primitive types, or void.  You can access the various primitive type
     * classes through java.lang.Boolean.TYPE, java.lang.Integer.TYPE, etc.,
     * or through boolean.class, int.class, etc.
     *
     * @return whether this class is a primitive type
     * @see Boolean#TYPE
     * @see Byte#TYPE
     * @see Character#TYPE
     * @see Short#TYPE
     * @see Integer#TYPE
     * @see Long#TYPE
     * @see Float#TYPE
     * @see Double#TYPE
     * @see Void#TYPE
     * @since 1.1
     */
    public boolean isPrimitive() {
	return kind == PRIMITIVE;
    }
    
    /**
     * Get the name of this class, separated by dots for package separators.
     * Primitive types and arrays are encoded as:
     * <pre>
     * boolean             Z
     * byte                B
     * char                C
     * short               S
     * int                 I
     * long                J
     * float               F
     * double              D
     * void                V
     * array type          [<em>element type</em>
     * class or interface, alone: &lt;dotted name&gt;
     * class or interface, as element type: L&lt;dotten name&gt;;
     *
     * @return the name of this class
     */
    public String getName() {
        // lazily initialize. There's no race as the same information is
        // always returned - an interned string instance
        if (name == null){
	    name = LibraryImports.nameForClass(this.vmType);
	}
        return name;
    }

    /**
     * Get the ClassLoader that loaded this class.  If it was loaded by the
     * bootstrap classloader, this method will return null. 
     * If there is a security
     * manager, and the caller's class loader does not match the requested
     * one, a security check of <code>RuntimePermission("GetClassLoader")</code>
     * must first succeed. Primitive types and void return null.
     *
     * @return the ClassLoader that loaded this class
     * @throws SecurityException if the security check fails
     * @see ClassLoader
     * @see RuntimePermission
     */
    public ClassLoader getClassLoader() {
        // Check some common cases.
        if (isPrimitive())
            return null;
        // don't do this optimisation yet as we want to ensure the base
        // mechanism is working correctly
         String name = getName();
         if (name.startsWith("java.") || name.startsWith("gnu.java."))
             return null;
        
        ClassLoader loader = LibraryImports.getClassLoader(vmType);
	// Other parts of the system assume that the bootstrap loader
	// is represented by null.
	if (loader == VMClassLoader.bootClassLoader)
	    return null;

        // Check if we may get the classloader
        SecurityManager sm = System.getSecurityManager();
        if (sm != null) {
            // Get the calling class and classloader
            Class c = VMSecurityManager.getClassContext()[1];
            ClassLoader cl = c.getClassLoader();
            if (cl != null && cl != ClassLoader.systemClassLoader)
                sm.checkPermission(new RuntimePermission("getClassLoader"));
        }
        return loader;
    }


    static volatile Class objectClass;
    static private Class getObjectClass() {
        if (objectClass == null)
            objectClass = Object.class;
        return objectClass;
    }
    
    /**
     * Get the direct superclass of this class.  If this is an interface,
     * Object, a primitive type, or void, it will return null. If this is an
     * array type, it will return Object.
     *
     * @return the direct superclass of this class
     */
    public Class getSuperclass(){
        if (vmType == null)
            throw new Error("Holy cow! - no vmType!");

        if (isInterface() || isPrimitive() || this == getObjectClass() ||
            this == Void.class)
            return null;
        if (isArray())
            return getObjectClass();
	return LibraryImports.getSuperclass(vmType);
    }
    
    /**
     * Returns the <code>Package</code> in which this class is defined
     * Returns null when this information is not available from the
     * classloader of this class or when the classloader of this class
     * is null.
     *
     * @return the package for this class, if it is available
     * @since 1.2
     */
    public Package getPackage() {
        ClassLoader cl = getClassLoader();
        if (cl != null)
            return cl.getPackage(getPackagePortion(getName()));
        return null;
    }

  /**
   * Strip the last portion of the name (after the last dot).
   *
   * @param name the name to get package of
   * @return the package name, or "" if no package
   */
    private static String getPackagePortion(String name) {
        int lastInd = name.lastIndexOf('.');
        if (lastInd == -1)
            return "";
        return name.substring(0, lastInd);
    }
    
    /**
     * Get the interfaces this class <EM>directly</EM> implements, in the
     * order that they were declared. This returns an empty array, not null,
     * for Object, primitives, void, and classes or interfaces with no direct
     * superinterface. Array types return Cloneable and Serializable.
     *
     * @return the interfaces this class directly implements
     */
    public native Class[] getInterfaces();
    
    /**
     * If this is an array, get the Class representing the type of array.
     * Examples: "[[Ljava.lang.String;" would return "[Ljava.lang.String;", and
     * calling getComponentType on that would give "java.lang.String".  If
     * this is not an array, returns null.
     *
     * @return the array type of this class, or null
     * @see Array
     * @since 1.1
     */
    public Class getComponentType() {
        if (isArray())
	    return LibraryImports.getComponentClass(vmType);
        return null;
    }
    
    /**
     * Get the modifiers of this class.  These can be decoded using Modifier,
     * and is limited to one of public, protected, or private, and any of
     * final, static, abstract, or interface. An array class has the same
     * public, protected, or private modifier as its component type, and is
     * marked final but not an interface. Primitive types and void are marked
     * public and final, but not an interface.
     *
     * @return the modifiers of this class
     * @see Modifer
     * @since 1.1
     */
    public int getModifiers(){
	return modifiers;
    }
    
    /**
     * Get the signers of this class. This returns null if there are no signers,
     * such as for primitive types or void.
     *
     * @return the signers of this class
     * @since 1.1
     */
    public Object[] getSigners()
    {
	return signers;
    }
    
    /**
     * Set the signers of this class.
     *
     * @param signers the signers of this class
     */
    void setSigners(Object[] signers)
    {
	this.signers = signers;
    }
    
    /**
     * If this is a nested or inner class, return the class that declared it.
     * If not, return null.
     *
     * @return the declaring class of this class
     * @since 1.1
     */
    public native Class getDeclaringClass();
    
    /**
     * Get all the public member classes and interfaces declared in this
     * class or inherited from superclasses. This returns an array of length
     * 0 if there are no member classes, including for primitive types. A
     * security check may be performed, with
     * <code>checkMemberAccess(this, Member.PUBLIC)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @return all public member classes in this class
     * @throws SecurityException if the security check fails
     * @since 1.1
     */
    public native Class[] getClasses();

   /**
     * Get all the public fields declared in this class or inherited from
     * superclasses. This returns an array of length 0 if there are no fields,
     * including for primitive types. This does not return the implicit length
     * field of arrays. A security check may be performed, with
     * <code>checkMemberAccess(this, Member.PUBLIC)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @return all public fields in this class
     * @throws SecurityException if the security check fails
     * @since 1.1
     */
    public Field[] getFields() {
	if (publicFields == null) {
	    publicFields = LibraryImports.getFields(vmType, this);
        }
	return publicFields;
    }

    /**
     * Get all the public methods declared in this class or inherited from
     * superclasses. This returns an array of length 0 if there are no methods,
     * including for primitive types. This does include the implicit methods of
     * arrays and interfaces which mirror methods of Object, nor does it
     * include constructors or the class initialization methods. The Virtual
     * Machine allows multiple methods with the same signature but differing
     * return types; all such methods are in the returned array. A security
     * check may be performed, with
     * <code>checkMemberAccess(this, Member.PUBLIC)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @return all public methods in this class
     * @throws SecurityException if the security check fails
     * @since 1.1
     */
    public Method[] getMethods(){
	if(publicMethods == null)
	    _initPublicMethods();
	return publicMethods;
    }
    
    /**
     * Get all the public constructors of this class. This returns an array of
     * length 0 if there are no constructors, including for primitive types,
     * arrays, and interfaces. It does, however, include the default
     * constructor if one was supplied by the compiler. A security check may
     * be performed, with <code>checkMemberAccess(this, Member.PUBLIC)</code>
     * as well as <code>checkPackageAccess</code> both having to succeed.
     *
     * @return all public constructors in this class
     * @throws SecurityException if the security check fails
     * @since 1.1
     */
    public Constructor[] getConstructors(){
	if(publicConstructors == null)
	    _initPublicConstructors();
	return publicConstructors;
    }
    
    /**
     * Get a public field declared or inherited in this class, where name is
     * its simple name. If the class contains multiple accessible fields by
     * that name, an arbitrary one is returned. The implicit length field of
     * arrays is not available. A security check may be performed, with
     * <code>checkMemberAccess(this, Member.PUBLIC)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @param name the name of the field
     * @return the field
     * @throws NoSuchFieldException if the field does not exist
     * @throws SecurityException if the security check fails
     * @see #getFields()
     * @since 1.1
     */
    public Field getField(String name) throws NoSuchFieldException {
	Field[] f = getFields();
	for(int i = 0; i < f.length; i++)
	    if (f[i].getName().equals(name))
		return f[i];
	throw new NoSuchFieldException();
    }
    
    /**
     * Get a public method declared or inherited in this class, where name is
     * its simple name. The implicit methods of Object are not available from
     * arrays or interfaces.  Constructors (named "<init>" in the class file)
     * and class initializers (name "<clinit>") are not available.  The Virtual
     * Machine allows multiple methods with the same signature but differing
     * return types, and the class can inherit multiple methods of the same
     * return type; in such a case the most specific return types are favored,
     * then the final choice is arbitrary. If the method takes no argument, an
     * array of zero elements and null are equivalent for the types argument.
     * A security check may be performed, with
     * <code>checkMemberAccess(this, Member.PUBLIC)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @param name the name of the method
     * @param types the type of each parameter
     * @return the method
     * @throws NoSuchMethodException if the method does not exist
     * @throws SecurityException if the security check fails
     * @see #getMethods()
     * @since 1.1
     */
    public Method getMethod(String name, Class[] args)
        throws NoSuchMethodException{
	if (publicMethods == null)
	    _initPublicMethods();

	MethodKey mkey = 
	    new MethodKey(name, args);
	Object foundEntry = methodHashMap.get(mkey);
	if (foundEntry != null){
	    /* it's faster to look straight in the hash table, but 
	     * we then have to check the modifier */
	    if (Modifier.isPublic(((Method)foundEntry).getModifiers()))
		return (Method) foundEntry;
	} 
	
	throw new NoSuchMethodException(formatSignature(name, args));	       
    }


    private String formatSignature(String name, Class[] args) {
        StringBuffer buf = new StringBuffer(32);
        buf.append(name);
        buf.append('(');
        int arglen = args == null ? 0 : args.length;
        for (int i = 0; i < arglen; i++) {
            buf.append(typeName(args[i].getName()));
            if (i < arglen-1)
                buf.append(", ");
        }
        buf.append(')');
        return buf.toString();
    }

    /**
     * Convert a fully qualified typename a.b.c to the simple name c 
     * (used to get the name of a constructor)
     */
    private String simpleTypeName(String name) {
        int lastDot = name.lastIndexOf('.');
        return name.substring(lastDot+1);
    }

    /**
     * Convert a type name, as returned by Class.getName() into the
     * expected readable form eg I -> int, [I -> int[], 
     * [Lfoo.bar.Baz; -> foo.bar.Baz[]
     */
    private String typeName(String name) {
        if (name.charAt(0) == '[') // array
            return arrayComponentName(name.substring(1)) + "[]";
        else
            return name;
    }

    private String arrayComponentName(String name) {
        if (name.charAt(0) == '[') // array
            return arrayComponentName(name.substring(1)) + "[]";

        if (name.length() == 1) { // primitive encoded in an array type
            switch( name.charAt(0)) {
            case 'Z': return "boolean";
            case 'B': return "byte";
            case 'C': return "char";
            case 'D': return "double";
            case 'F': return "float";
            case 'I': return "int";
            case 'J': return "long";
            case 'S': return "short";
            default: throw new Error("Unexpected type name " + name);
            }
        }
        // else it is an array component name L?????; 
        if (name.charAt(0) == 'L') {
            // strip the L and the ; 
            return name.substring(1, name.length()-1);
        }
        else {
            throw new Error("Unexpected array component name " + name);
        }
    }
        

    /**
     * Get a public constructor declared in this class. If the constructor takes
     * no argument, an array of zero elements and null are equivalent for the
     * types argument. A security check may be performed, with
     * <code>checkMemberAccess(this, Member.PUBLIC)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @param types the type of each parameter
     * @return the constructor
     * @throws NoSuchMethodException if the constructor does not exist
     * @throws SecurityException if the security check fails
     * @see #getConstructors()
     * @since 1.1
     */
    public Constructor getConstructor(Class[] args)
        throws NoSuchMethodException {
	if (publicConstructors == null)
	    _initPublicConstructors();
	MethodKey mkey = 
	    new MethodKey(constructorName, args);
	Object foundEntry = constructorHashMap.get(mkey);
	if (foundEntry != null) {
	    /* it's faster to look straight in the hash table, but 
	     * we then have to check the modifier */
	    if (Modifier.isPublic(((Constructor)foundEntry).getModifiers()))
		return (Constructor) foundEntry;
	}

	throw new NoSuchMethodException("(constructor) " + formatSignature(simpleTypeName(this.getName()), args));
    }
    
    /**
     * Get all the declared member classes and interfaces in this class, but
     * not those inherited from superclasses. This returns an array of length
     * 0 if there are no member classes, including for primitive types. A
     * security check may be performed, with
     * <code>checkMemberAccess(this, Member.DECLARED)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @return all declared member classes in this class
     * @throws SecurityException if the security check fails
     * @since 1.1
     */
    public native Class[] getDeclaredClasses();
    
    /**
     * Get all the declared fields in this class, but not those inherited from
     * superclasses. This returns an array of length 0 if there are no fields,
     * including for primitive types. This does not return the implicit length
     * field of arrays. A security check may be performed, with
     * <code>checkMemberAccess(this, Member.DECLARED)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @return all declared fields in this class
     * @throws SecurityException if the security check fails
     * @since 1.1
     */
    public Field[] getDeclaredFields() {
        if (declaredFields == null){
            declaredFields = LibraryImports.getDeclaredFields(vmType, this);
        }
        return declaredFields;
    }
    
    /**
     * Get all the declared methods in this class, but not those inherited from
     * superclasses. This returns an array of length 0 if there are no methods,
     * including for primitive types. This does include the implicit methods of
     * arrays and interfaces which mirror methods of Object, nor does it
     * include constructors or the class initialization methods. The Virtual
     * Machine allows multiple methods with the same signature but differing
     * return types; all such methods are in the returned array. A security
     * check may be performed, with
     * <code>checkMemberAccess(this, Member.DECLARED)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @return all declared methods in this class
     * @throws SecurityException if the security check fails
     * @since 1.1
     */
    public Method[] getDeclaredMethods(){
	if(declaredMethods == null){
	    _initDeclaredMethods();
	}
	return declaredMethods;
    }
    
    /**
     * Get all the declared constructors of this class. This returns an array of
     * length 0 if there are no constructors, including for primitive types,
     * arrays, and interfaces. It does, however, include the default
     * constructor if one was supplied by the compiler. A security check may
     * be performed, with <code>checkMemberAccess(this, Member.DECLARED)</code>
     * as well as <code>checkPackageAccess</code> both having to succeed.
     *
     * @return all constructors in this class
     * @throws SecurityException if the security check fails
     * @since 1.1
     */
    public Constructor[] getDeclaredConstructors(){
	if(declaredConstructors == null) {
	    _initDeclaredConstructors();
	}
	return declaredConstructors;
    }
    
    /**
     * Get a field declared in this class, where name is its simple name. The
     * implicit length field of arrays is not available. A security check may
     * be performed, with <code>checkMemberAccess(this, Member.DECLARED)</code>
     * as well as <code>checkPackageAccess</code> both having to succeed.
     *
     * @param name the name of the field
     * @return the field
     * @throws NoSuchFieldException if the field does not exist
     * @throws SecurityException if the security check fails
     * @see #getDeclaredFields()
     * @since 1.1
     */
    public Field getDeclaredField(String name) throws NoSuchFieldException {
	Field[] f = getDeclaredFields();
	//FIXME: linear search
	for(int i = 0; i < f.length; i++)
	    if (f[i].getName().equals(name))
		return f[i];
	throw new NoSuchFieldException(name);
    }
    
    /**
     * Get a method declared in this class, where name is its simple name. The
     * implicit methods of Object are not available from arrays or interfaces.
     * Constructors (named "<init>" in the class file) and class initializers
     * (name "<clinit>") are not available.  The Virtual Machine allows
     * multiple methods with the same signature but differing return types; in
     * such a case the most specific return types are favored, then the final
     * choice is arbitrary. If the method takes no argument, an array of zero
     * elements and null are equivalent for the types argument. A security
     * check may be performed, with
     * <code>checkMemberAccess(this, Member.DECLARED)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @param name the name of the method
     * @param types the type of each parameter
     * @return the method
     * @throws NoSuchMethodException if the method does not exist
     * @throws SecurityException if the security check fails
     * @see #getDeclaredMethods()
     * @since 1.1
     */
    public Method getDeclaredMethod(String name, Class[] args)
        throws NoSuchMethodException{
	if (declaredMethods == null)
	    _initDeclaredMethods();

	MethodKey mkey = 
	    new MethodKey(name, args);
	Object foundEntry = methodHashMap.get(mkey);
	if (foundEntry != null){
		return (Method) foundEntry;
	} 
	
	throw new NoSuchMethodException(formatSignature(name, args));
    }
    
    /**
     * Get a constructor declared in this class. If the constructor takes no
     * argument, an array of zero elements and null are equivalent for the
     * types argument. A security check may be performed, with
     * <code>checkMemberAccess(this, Member.DECLARED)</code> as well as
     * <code>checkPackageAccess</code> both having to succeed.
     *
     * @param types the type of each parameter
     * @return the constructor
     * @throws NoSuchMethodException if the constructor does not exist
     * @throws SecurityException if the security check fails
     * @see #getDeclaredConstructors()
     * @since 1.1
     */
    public Constructor getDeclaredConstructor(Class[] args)
	throws NoSuchMethodException
    {
	if(declaredConstructors == null)
	    _initDeclaredConstructors();

	MethodKey mkey = 
	    new MethodKey(constructorName, args);
	Object foundEntry = constructorHashMap.get(mkey);
	if (foundEntry != null) {
	    return (Constructor) foundEntry;
        }

	throw new NoSuchMethodException("(constructor) " + formatSignature(simpleTypeName(this.getName()), args));
    }


    /**
     * Get a resource using this class's package using the
     * getClassLoader().getResourceAsStream() method.  If this class was loaded
     * using the system classloader, ClassLoader.getSystemResource() is used
     * instead.
     *
     * <p>If the name you supply is absolute (it starts with a <code>/</code>),
     * then it is passed on to getResource() as is.  If it is relative, the
     * package name is prepended, and <code>.</code>'s are replaced with
     * <code>/</code>.
     *
     * <p>The URL returned is system- and classloader-dependent, and could
     * change across implementations.
     *
     * @param name the name of the resource, generally a path
     * @return an InputStream with the contents of the resource in it, or null
     * @throws NullPointerException if name is null
     * @since 1.1
     */
    public InputStream getResourceAsStream(String name)
        {
            if (name.length() > 0 && name.charAt(0) != '/')
                name = getPackagePortion(getName()).replace('.','/')
                    + "/" + name;
            ClassLoader c = getClassLoader();
            if (c == null)
                return ClassLoader.getSystemResourceAsStream(name);
            return c.getResourceAsStream(name);
        }
    
    /**
     * Get a resource URL using this class's package using the
     * getClassLoader().getResource() method.  If this class was loaded using
     * the system classloader, ClassLoader.getSystemResource() is used instead.
     *
     * <p>If the name you supply is absolute (it starts with a <code>/</code>),
     * then it is passed on to getResource() as is.  If it is relative, the
     * package name is prepended, and <code>.</code>'s are replaced with
     * <code>/</code>.
     *
     * <p>The URL returned is system- and classloader-dependent, and could
     * change across implementations.
     *
     * @param name the name of the resource, generally a path
     * @return the URL to the resource
     * @throws NullPointerException if name is null
     * @since 1.1
     */
    public URL getResource(String name)
        {
            if(name.length() > 0 && name.charAt(0) != '/')
                name = getPackagePortion(getName()).replace('.','/')
                    + "/" + name;
            ClassLoader c = getClassLoader();
            if (c == null)
                return ClassLoader.getSystemResource(name);
            return c.getResource(name);
        }
    
    /**
     * Returns the protection domain of this class. If the classloader did not
     * record the protection domain when creating this class the unknown
     * protection domain is returned which has a <code>null</code> code source
     * and all permissions. A security check may be performed, with
     * <code>RuntimePermission("getProtectionDomain")</code>.
     *
     * @return the protection domain
     * @throws SecurityException if the security check fails
     * @see RuntimePermission
     * @since 1.2
     */
    public ProtectionDomain getProtectionDomain()
        {
            SecurityManager sm = System.getSecurityManager();
            if (sm != null)
                sm.checkPermission(new RuntimePermission("getProtectionDomain"));
            
            return pd == null ? unknownProtectionDomain : pd;
        }
    
    /**
     * Returns the desired assertion status of this class, if it were to be
     * initialized at this moment. The class assertion status, if set, is
     * returned; the backup is the default package status; then if there is
     * a class loader, that default is returned; and finally the system default
     * is returned. This method seldom needs calling in user code, but exists
     * for compilers to implement the assert statement. Note that there is no
     * guarantee that the result of this method matches the class's actual
     * assertion status.
     *
     * @return the desired assertion status
     * @see ClassLoader#setClassAssertionStatus(String, boolean)
     * @see ClassLoader#setPackageAssertionStatus(String, boolean)
     * @see ClassLoader#setDefaultAssertionStatus(boolean)
     * @since 1.4
     */
    public boolean desiredAssertionStatus()
        {
            ClassLoader c = getClassLoader();
            Object status;
            if (c == null)
                return VMClassLoader.defaultAssertionStatus();
            if (c.classAssertionStatus != null)
                synchronized (c)
                {
                    status = c.classAssertionStatus.get(getName());
                    if (status != null)
                        return status.equals(Boolean.TRUE);
                }
            else
            {
                status = ClassLoader.systemClassAssertionStatus.get(getName());
                if (status != null)
                    return status.equals(Boolean.TRUE);
            }
            if (c.packageAssertionStatus != null)
                synchronized (c)
                {
                    String name = getPackagePortion(getName());
                    if ("".equals(name))
                        status = c.packageAssertionStatus.get(null);
                    else
                        do
                        {
                            status = c.packageAssertionStatus.get(name);
                            name = getPackagePortion(name);
                        }
                        while (! "".equals(name) && status == null);
                    if (status != null)
                        return status.equals(Boolean.TRUE);
                }
            else
            {
                String name = getPackagePortion(getName());
                if ("".equals(name))
                    status = ClassLoader.systemPackageAssertionStatus.get(null);
                else
                    do
                    {
                        status = ClassLoader.systemPackageAssertionStatus.get(name);
                        name = getPackagePortion(name);
                    }
                    while (! "".equals(name) && status == null);
                if (status != null)
                    return status.equals(Boolean.TRUE);
            }
            return c.defaultAssertionStatus;
        }


    
    /* methods to deal with j.l.r.Method and j.l.r.Constructor */
    private void _initDeclaredMethods(){
        int nb = 0;
        for(Iterator iter = _getMethodHashMapIterator(); iter.hasNext();){
            Method m = (Method) iter.next();
            if(m.getDeclaringClass() ==this)
                nb++;
        }
        Method[] methods = new Method[nb];
        nb = 0;
        for(Iterator iter = _getMethodHashMapIterator(); iter.hasNext();){
            Method m = (Method) iter.next();
            if(m.getDeclaringClass() ==this)
                methods[nb++] = m;
        }
        declaredMethods = methods;
    }
    private void _initPublicMethods(){
        int nb = 0;
        for(Iterator iter = _getMethodHashMapIterator(); iter.hasNext();){
            Method m = (Method) iter.next();
            if(Modifier.isPublic(m.getModifiers()))
                nb++;
        }
        Method[] methods = new Method[nb];
        nb = 0;
        for(Iterator iter = _getMethodHashMapIterator(); iter.hasNext();){
            Method m = (Method) iter.next();
            if(Modifier.isPublic(m.getModifiers()))
                methods[nb++] = m;
        }
        publicMethods = methods;
    }
    private void _initDeclaredConstructors(){
        int nb = 0;
        for(Iterator iter = _getConstructorHashMapIterator(); iter.hasNext();){
            Constructor m = (Constructor) iter.next();
            nb++;
        }
        Constructor[] constructors = new Constructor[nb];
        nb = 0;
        for(Iterator iter = _getConstructorHashMapIterator(); iter.hasNext();){
            Constructor m = (Constructor) iter.next();
            constructors[nb++] = m;
	}
        declaredConstructors = constructors;
    }
    private void _initPublicConstructors(){
        int nb = 0;
        for(Iterator iter = _getConstructorHashMapIterator(); iter.hasNext();){
            Constructor m = (Constructor) iter.next();
            if (Modifier.isPublic(m.getModifiers()))
                nb++;
        }
        Constructor[] constructors = new Constructor[nb];
        nb = 0;
        for(Iterator iter = _getConstructorHashMapIterator(); iter.hasNext();){
            Constructor m = (Constructor) iter.next();
            if(Modifier.isPublic(m.getModifiers()))
                constructors[nb++] = m;
        }
        publicConstructors = constructors;
    }

    /* returns an iterator on the hash map of methods ; initializes the
     * the method hash map if needed */
    private Iterator _getMethodHashMapIterator(){
	if(methodHashMap == null)
	    _initMethodHash();
	return methodHashMap.values().iterator();
    }
    private Iterator _getConstructorHashMapIterator(){
	if(constructorHashMap == null)
	    _initConstructorHash();
	return constructorHashMap.values().iterator();
    }

    /**
     * Initializes the hash table that stores methods provided by this class,
     * either locally declared or inherited.
     * Overriden methods are found only once in the hash table : we only
     * store the overriding one.
     */    
    private void _initMethodHash(){
 	Class parent = getSuperclass();

	/* if we have a super class, get its method hash map */
	if(parent == null)
	    methodHashMap = new HashMap();
	else
	    methodHashMap = parent.getMethodHashMap();	

	/* get our declared methods */
	Method[] myDeclaredMethods = 
	    LibraryImports.getDeclaredMethods(vmType, this);

	/* merge the methods in the hash table 
	 * Since the hash key works on method name and args, this will
	 * take care of overriding at the same time */
	for(int i=0; i<myDeclaredMethods.length; i++){
	    MethodKey mkey = 
		new MethodKey(myDeclaredMethods[i].getName(),
			      myDeclaredMethods[i].getParameterTypes());
	    methodHashMap.put(mkey, myDeclaredMethods[i]);
	}
    }

    private void _initConstructorHash(){

	constructorHashMap = new HashMap();

	/* get our declared constructors */
	Constructor[] myDeclaredConstructors = 
	    LibraryImports.getDeclaredConstructors(vmType, this);

	/* put the constructors in the hash table 
	 * Since the hash key works on constructor name and args, this will
	 * take care of overriding at the same time */
	for(int i=0; i<myDeclaredConstructors.length; i++){
	    MethodKey mkey = 
		new MethodKey(constructorName,
			      myDeclaredConstructors[i].getParameterTypes());
	    constructorHashMap.put(mkey, myDeclaredConstructors[i]);
	}
    }
	

    /**
     * @return a copy of the method hash table from this class ; this is 
     * used by classes to get learn what methods they inherit from their
     * super class
     * @see initMethodHash()
     */
    private HashMap getMethodHashMap(){
	if(methodHashMap == null)
	    _initMethodHash();
	return (HashMap) methodHashMap.clone();
    }
    
    /**
     * The key we use to hash methods in the method hash table
     */
    private static class MethodKey{
        static final Class[] EMPTY_ARRAY = new Class[0];
	Class[] args;
	String name;
	
	public MethodKey(String name, Class[] args){
	    this.name = name;
	    this.args = (args != null ? args : EMPTY_ARRAY);
	}

	public boolean equals(Object o){
	    return (hashCode() == o.hashCode());
	}

	public int hashCode(){
	    int c = name.hashCode();
            // methods/constructors that have certain parameter combinations
            // will yield the same hash. For example for any method of the
            // form meth(X x1, X x2) will have the same hash as the no arg
            // constructor. Similarly meth(X x, Y y) and meth(Y y, X x) will
            // has the same. So we shift the hash of the arg type by the 
            // position of the arg in an attempt to fix this.
            // This hash may still not be unique - FIXME
	    for(int i=0; i<args.length; i++)
		c ^= (args[i].hashCode() << i);
	    return c;
	}

        public String toString() {
            return name + ":" + args.length + "@" + hashCode();
        }
    }
    
} // class Class
